/* 
 * ****************************************************************************
 * RP6 ROBOT SYSTEM - ROBOT BASE EXAMPLES
 * ****************************************************************************
 * Example: Serial Interface Example - Question and Answer
 * Author(s): Dominik S. Herwald
 * ****************************************************************************
 * Description:
 * This sample program shows how to use the RP6Library functions to receive
 * data with the serial interface. It is a "Question and Answer" program. 
 * It is a bit longer, but don't be afraid - the next programs are a lot
 * shorter again. Also have a look at the uncommented version of this program!
 *
 * ############################################################################
 * The Robot does NOT move in this example! You can simply put it on a table
 * next to your PC and you should connect it to the PC via the USB Interface!
 * ############################################################################
 * ****************************************************************************
 */

/*****************************************************************************/
// Includes:

#include "RP6RobotBaseLib.h" 	// The RP6 Robot Base Library.
								// Always needs to be included!


/*****************************************************************************/
// Chaselight

/**
 * This is a small chaselight function. This has nothing to do with the 
 * UART, it is just a demo application. 
 * Btw. this is nearly the same code as in example 1 - just without the comments!
 * And this one has a small "blanking" effect.
 */
void runningLight(void)
{
	static uint8_t runningLight = 1; // static variable used to store 	
								   // running light value. A static variable 
								   // does NOT "forget" about its value when
								   // the function ends. It is like a global
								   // variable, but only accessible within this
								   // Function!
								   // runningLight is initialized with 1 
								   // ONCE only!
								   
	// The chaselight code fragment you already know from example 1:
	setLEDs(runningLight); 	
	runningLight <<= 1;
	if(runningLight > 32)
		runningLight = 1; 	
	mSleep(100); 
	
	// A small blanking effekt...
	// If StatusLED3 or StatusLED6 are on, 
	// turn off all LEDs for 200ms:
	if(statusLEDs.LED3 || statusLEDs.LED6 )
	{
		setLEDs(0b000000);
		mSleep(200);
	}
}

/**
 * Another demo function that will be called as a response to one of your answers.
 * It uses the "writeStringLength" function to output some text - well this is hard to 
 * describe - just look at the output on the serial interface!
 */
void outputTextStuff(void)
{
	uint8_t x, y; // define TWO local variables x and y. (In comparison to the static
				  // variable in the function above, these DO "forget" about their values
				  // when this function ends)
				 
	for(x = 0; x < 53; x++)
	{
		writeStringLength("1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",10,x);
		writeChar(' ');
		for(y = 0; y < x; y++) // output x times "#" on the serial interface...
			writeChar('#');
		mSleep(200); 
		writeChar('\n');
	}
}

/*****************************************************************************/
// Question game

// A global variable that stores the current question:
int8_t question = 0;

/**
 * Here we ask a question and receive the answer from the user. 
 * The function uses some of the reception routines of the RP6Library.
 * It is called from the main loop (s. below).
 */
void askAQuestion(void)
{
	uint8_t bytesToReceive = 1; // By default, we only want to receive 1 character
	
	// Ask a question (the variable "question" contains the number of the current question):
	switch(question)
	{
		case 0: // Question 0
			writeString_P("What's your name? (Enter 8 characters please)"); 
			bytesToReceive = 8;  // here we want to receive 8 characters
								 // and not only the default 1 character
		break;
		case 1: // Question 1 ...
			writeString_P("How old are you? (Enter 2 characters please)"); 
			bytesToReceive = 2; // receive 2 characters
		break;
		case 2: // 2 ...
			writeString_P("Do you want to see some Text output? (\"y\" or \"n\")"); 
			// bytesToReceive has the value 1 by default, so we don't need to 
			// set it to 1 to receive one character.
		break;
		case 3: 
			writeString_P("Do you want to see blinking LEDs? (\"y\" or \"n\")"); 
		break;
		case 4: 
			writeString_P("Do you want to do all this again? (\"y\" or \"n\")"); 
		break;
	}
	
	char receiveBuffer[bytesToReceive]; // our reception buffer
	// Receive the answer:
	receiveBytes(bytesToReceive); // Setup reception with number of bytes to receive
	waitUntilReceptionComplete(); // wait until everythig is received...
								  // This function waits FOREVER - it will not stop to
								  // wait for _all_ requested characters to be received!
								  
	// ... and copy the received data to our local reception buffer:
	copyReceivedBytesToBuffer(&receiveBuffer[0]);
	
	// "&receiveBuffer[0]" is a pointer to the position 0 in the array
	
	// Do something with the received answer:
	switch(question) // select the question...
	{
		case 0: // Question 0: What's your name?
			writeString_P("Hello \"");
			writeStringLength(&receiveBuffer[0],bytesToReceive,0); // Output the received data as a String
			writeString_P("\" !\n");
		break;
		case 1: // Your age?
			writeString_P("So your age is: \""); 
			writeStringLength(&receiveBuffer[0],bytesToReceive,0);
			writeString_P("\" ! Guess what? Nobody cares... ;-)\n");
		break;
		case 2: // Text output?
			// Check if the answer was "y":
			if(receiveBuffer[0] == 'y') // Answer is 'y' ?
			{
				// Output some text. Here we use the writeStringLength
				// function to demonstrate a bit better how this one works. 
				writeString_P("Here we go!\n");
				outputTextStuff(); 	// Call output function
				writeString_P("\n\nThat's it!\n\n"); 
			}
			else // No the answer was not 'y' ... most likely it was 'n', but it does
			     // not have to be 'n' - we don't need to care about that.
				writeString_P("OK then... let's move on!\n"); 
		break;
		case 3: // Chaselight?
			if(receiveBuffer[0] == 'y') // Answer is 'y' ?
			{
				writeString_P("Now you get the ultimative chase light!\n");
				uint8_t i = 0;
				for(;i < 36; i++) // run the chaselight function 36 times
					runningLight();
				writeString_P("Oh ... it's over!\n");
			}
			else
				writeString_P("Well, if you don't want that... :*( \n"); 
		break;
		case 4: // Again?
			if(receiveBuffer[0] == 'y') // Answer was 'y' ?
			{
				writeString_P("Great! Here we go again...\n"); 
				question = -1; // If "question" is set to -1, the question/answer game
							   // will jump back to the first question as the variable "question"
							   // is incremented at the end of this function (s. line // Next question...)
							   // it gets the value 0
			}
			else
			{
				writeString_P("OK good bye! I will rest in an infinite running light loop\n");
				writeString_P("until someone resets me!\n"); 
				question = -2; // If "question" is set to -2, the increment at the end
							   // of this function only sets it to -1 and thus the main
							   // loop will quit the question answer game as it checks for this!
							   // (s. below)
			}
		break;
	}
	
	question++; // Next question...
}


/*****************************************************************************/
// Main function - The program starts here:

int main(void)
{
	initRobotBase(); // Always call this first! The Processor will not work
					 // correctly otherwise.

	setLEDs(0b111111); // Turn all LEDs on
	mSleep(500);       // delay 500ms
	setLEDs(0b000000); // All LEDs off

	// Write a message to UART:
	writeString_P("\nHello! My name is RP6!\n\n");
	
	// ---------------------------------------
	// Main loop:
	while(true)
	{
		if(question != -1) // Ask a question? This gets false if "question" has the value -1.
			askAQuestion(); // Yes!
		else              // No we are done! (question = -1 ...)
			runningLight(); // END! Show chaselight, forever!
	}
	// End of main loop
	// ---------------------------------------

	return 0;
}

/*****************************************************************************/
// EOF
